"""
harness stuff (support)
"""

import os
import shutil
import sys
import unittest
import re
import time

#
# globals
#
class Global:
    # tmp directories where the tests will be run
    _isns_test_base_path = '/tmp/isns-test'
    _isns_test_path = '/tmp/isns-test/test'
    # various attributes
    _isns_test_dump = None
    _isns_test_data = None
    _isns_seq = 0
    _isns_servers = []
    _isns_server_config_data = {}
    _isns_stage_name = None
    _isns_stage_names = []

    _isns_ignore_tags = ['0004', '0603v']

    # where the tests are relative to where we start them from
    # default to "current directory"
    test_dir = '.'
    # for locating binaries under test_dir (FIXME: should be a command-line option?)
    isns_bin_dir = '../builddir'

    verbosity = 1
    security = False
    debug = False

def dprint(*args):
    """
    Print a debug message if in debug mode
    """
    if Global.debug:
        print('DEBUG: ', file=sys.stderr, end='')
        for arg in args:
            print(arg, file=sys.stderr, end='')
        print('', file=sys.stderr)

def vprint(*args):
    """
    Print a verbose message
    """
    if Global.verbosity > 0 and args:
        for arg in args:
            print(arg, end='')
        print('')

def get_bindir():
    """
    return the directory where binaries live
    """
    return '%s/%s' % (Global.test_dir, Global.isns_bin_dir)

def isns_stage(name, msg):
    """
    gaurd against duplicate stage names, since data files are named, based on them
    """
    if name in Global._isns_stage_names:
        print('internal error: duplicate stage name: %s\n' % name, file=sys.stderr)
        sys.exit(1)
    Global._isns_stage_name = name
    Global._isns_stage_names.append(name)
    vprint('*** Stage %s: %s ***' % (name, msg))

def set_up_test(test_name, security=None):
    Global._isns_test_path = '%s/%s' % (Global._isns_test_base_path, test_name)
    Global._isns_test_dump = '%s/dump' % Global._isns_test_path
    Global._isns_test_data = '%s/data/%s' % (Global.test_dir, test_name)

    if Global._isns_test_path.startswith('/tmp/'):
        if os.path.isdir(Global._isns_test_path):
            shutil.rmtree(Global._isns_test_path)

    if not os.path.isdir(Global._isns_test_path):
        os.makedirs(Global._isns_test_path, 0o755)
    if not os.path.isdir(Global._isns_test_dump):
        os.mkdir(Global._isns_test_dump, 0o755)

    if security is not None:
        dprint("Setting security=", security)
        Global.security = security

    # start clean
    Global._isns_seq = 0
    Global._isns_stage_name = None
    Global._isns_stage_names = []

    return Global.security

def run_cmd(cmd, output_save_file=None):
    """
    run specified command, waiting for and returning result
    """
    if Global.verbosity > 1:
        cmd_str = ' '.join(cmd)
        if output_save_file:
            cmd_str += ' >& %s' % output_save_file
        vprint(cmd_str)
    pid = os.fork()
    if pid < 0:
        print("Error: cannot fork!", flie=sys.stderr)
        sys.exit(1)
    if pid == 0:
        # the child
        if output_save_file:
            stdout_fileno = sys.stdout.fileno()
            stderr_fileno = sys.stderr.fileno()
            new_stdout = os.open(output_save_file, os.O_WRONLY|os.O_CREAT|os.O_TRUNC,
                                 mode=0o664)
            os.dup2(new_stdout, stdout_fileno)
            os.dup2(new_stdout, stderr_fileno)
        os.execvp(cmd[0], cmd)
        # not reached

    # the parent
    wpid, wstat = os.waitpid(pid, 0)
    if wstat != 0:
        dprint("exit status: (%d) %d" % (wstat, os.WEXITSTATUS(wstat)))
    return os.WEXITSTATUS(wstat)

def new_initArgParsers(self):
    """
    Add  some options to the normal unittest main options
    """
    global old_initArgParsers

    old_initArgParsers(self)
    self._main_parser.add_argument('-s', '--secure', dest='security',
            action='store_true',
            help='Enable security')
    self._main_parser.add_argument('-d', '--debug', dest='debug',
            action='store_true',
            help='Enable developer debugging')
    self._main_parser.add_argument('-l', '--list', dest='list',
            action='store_true',
            help='List tests and exit')
    self._main_parser.add_argument('-D', '--test-dir', dest='test_dir',
            default='.',
            help='Sets the test dir (default ".")')

def new_parseArgs(self, argv):
    """
    Gather globals from unittest main for local consumption
    """
    global old_parseArgs

    old_parseArgs(self, argv)
    Global.verbosity = self.verbosity
    Global.security = self.security
    Global.debug = self.debug
    Global.test_dir = self.test_dir
    dprint("found: debug=%s, verbosity=%d, security=%s, list=%s, test_dir='%s'" % \
           (Global.debug, Global.verbosity,
            Global.security, self.list, self.test_dir))
    if self.list:
        self.list_tests()

def setup_testProgram_overrides():
    """
    Add in special handling for a couple of the methods in TestProgram (main)
    so that we can add parameters and detect some globals we care about
    """
    global old_parseArgs, old_initArgParsers

    old_initArgParsers = unittest.TestProgram._initArgParsers
    unittest.TestProgram._initArgParsers = new_initArgParsers
    old_parseArgs = unittest.TestProgram.parseArgs
    unittest.TestProgram.parseArgs = new_parseArgs

def create_key(keyfile):
    """
    create a key file for security, if enabled
    """
    if Global.security:
        vprint('*** Creating key at %s' % keyfile)
        res = run_cmd(['./genkey', '-fsk', keyfile, '2048'], '%s.log' % keyfile)
        if res != 0:
            print('internal error: cannot general key (%d)' % res)
            sys.exit(1)
    return keyfile

class isns_config(object):
    name = ''
    def __init__(self, path):
        self.path = path
        self.dict = dict()

def build_config(from_config, to_config, local_config):
    """
    build a config file from the from_config, using local_config,
    and creating to_config
    """
    from_config_path = '%s/%s' % (Global.test_dir, from_config)
    vprint('*** Building %s -> %s' % (from_config_path, to_config))
    pat = re.compile(r'(\S+)(\s*=\s*)(\S+)')
    skip_pat = re.compile(r'@[A-Z_]*@')
    result = isns_config(to_config)
    with open(from_config_path, 'r') as from_fd:
        with open(to_config, 'a') as to_fd:
            for line in from_fd:
                res = pat.match(line.rstrip())
                if not res:
                    continue
                key_val = res.group(1)
                middle = res.group(2)
                val = res.group(3)
                if key_val in local_config:
                    val = local_config[key_val]
                    line = '%s%s%s\n' % (key_val, middle, val)
                elif skip_pat.search(line):
                    #dprint("Skipping unconfigured line!")
                    continue
                result.dict[key_val] = val
                to_fd.write(line)
    return result
        
def create_server(overrides=None):
    """
    Create a server configration instance
    """
    dprint("create_server: PWD=%s" % os.getcwd())
    handle = 'server%d' % Global._isns_seq
    Global._isns_seq += 1
    my_dir = '%s/%s' % (Global._isns_test_path, handle)
    if not os.path.isdir(my_dir):
        os.mkdir(my_dir, 0o755)
    if not os.path.isdir(Global._isns_test_dump):
        dprint("making dir %s for the server ???" % Global._isns_test_dump)
        os.mkdir(Global._isns_test_dump, 0o755)

    server_addr = '127.0.0.1:7770'

    server_config = '%s/config' % my_dir

    local_config = dict()
    local_config['SourceName'] = 'isns.%s' % handle
    local_config['Database'] = '%s/database' % my_dir
    local_config['BindAddress'] = server_addr
    local_config['PIDFile'] = '%s/pid' % my_dir
    local_config['ControlSocket'] = '%s/control' % my_dir
    local_config['Security'] = 1 if Global.security else 0
    local_config['AuthKeyFile'] = create_key('%s/auth_key' % my_dir)

    if overrides:
        dprint("overrides:", overrides)
        for key in overrides:
            value = overrides[key]
            dprint("overriding config[%s] with %s" % \
                   (key, value))
            local_config[key] = value

    return build_config('server.conf', server_config, local_config)

def create_client(server_config, client_address=None):
    """
    create a client, given a server config file
    """
    handle = 'client%d' % Global._isns_seq
    Global._isns_seq += 1
    my_dir = '%s/%s' % (Global._isns_test_path, handle)
    if not os.path.isdir(my_dir):
        os.mkdir(my_dir, 0o755)

    control_socket = server_config.dict['ControlSocket']

    if 'BindAddress' in server_config.dict:
        server_addr = server_config.dict['BindAddress']
    else:
        server_addr = '127.0.0.1:7770'

    client_config_path = '%s/config' % my_dir

    local_config = dict()
    local_config['SourceName'] = 'isns.%s' % handle
    local_config['AuthName'] = '%s.isns-test.eu' % handle
    local_config['ServerAddress'] = server_addr
    local_config['ControlSocket'] = control_socket
    if client_address:
        local_config['BindAddress'] = client_address
    local_config['server_config'] = server_config.path
    local_config['Security'] = 1 if Global.security else 0
    local_config['AuthKeyFile'] = create_key('%s/auth_key' % my_dir)
    local_config['ServerKeyFile'] = \
                   '%s.pub' % server_config.dict['AuthKeyFile']

    client_config = build_config('client.conf', client_config_path, local_config)

    Global._isns_server_config_data[client_config_path] = server_config.path

    return client_config

def get_logfile_from_config(config):
    return '%s/logfile' % os.path.dirname(config)

def isns_start_server(server_config):
    """
    start the isnsd server
    """
    if not os.path.isfile(server_config.path):
        print("internal error: no config file: %s" % server_config.path,
              file=sys.stderr)
        sys.exit(1)

    logfile = get_logfile_from_config(server_config.path)
    pidfile = server_config.dict['PIDFile']

    vprint('*** Starting server (logging to %s)' % logfile)

    cmd = ['%s/isnsd' % get_bindir(),
           '-c', server_config.path,
           '-f',
           '-d', 'all' ]
    if Global.verbosity > 0:
        cmd_str = '%s >& %s' % (' '.join(cmd), logfile)
        vprint(cmd_str)

    pid = os.fork()
    if pid < 0:
        print("internal error: cannot fork!", file=sys.stderr)
        sys.exit(1)

    if pid == 0:
        # the child -- run 'cmd 2>&1'
        stdout_fileno = sys.stdout.fileno()
        stderr_fileno = sys.stderr.fileno()
        new_stdout = os.open(logfile, os.O_WRONLY|os.O_CREAT|os.O_TRUNC,
                             mode=0o664)
        os.dup2(new_stdout, stdout_fileno)
        os.dup2(new_stdout, stderr_fileno)
        os.execvp(cmd[0], cmd)
        # not reached

    # the parent -- get child PID from file to be sure its ready
    dprint("parent: waiting for child PIDFile=%s (pid=%d)" % \
           (pidfile, pid))

    # wait for the PID file to appear for a bit, then get PID from it
    for retry in range(15):
        if os.path.isfile(pidfile):
            break
        time.sleep(1)
    if not os.path.isfile(pidfile):
        print("internal error: no pid file: %s" % pidfile, file=sys.stderr)
        sys.exit(1)
    with open(pidfile, 'r') as pid_fd:
        pid_from_file = int(pid_fd.readline().strip())
    vprint('*** Started server (pid=%d) ***' % pid_from_file)
    if pid_from_file in Global._isns_servers:
        print('internal error: duplicate server PID?', file=sys.stderr)
        sys.exit(1)
    Global._isns_servers.append(pid_from_file)
    return pid_from_file

def isns_stop_server(server_pid):
    """
    stop a single server with supplied PID, and remove said PID from server list
    """
    dprint('Killing isnsd server pid=%d' % server_pid)
    try:
        os.kill(server_pid, 15)
    except ProcessLookupError:
        print('warning: cannot send signal to server pid=%d: no such process' % \
              server_pid)
    if server_pid in Global._isns_servers:
        Global._isns_servers.remove(server_pid)

def isns_idle(sleep_amt):
    if Global.verbosity > 1:
        print('Snooze', end='')
        sys.stdout.flush()
    for secs in range(sleep_amt):
        if Global.verbosity > 1:
            print('.', end='')
            sys.stdout.flush()
        time.sleep(1)
    if Global.verbosity > 1:
        print('')
        
def isns_restart_server(server_config, server_pid):
    isns_stop_server(server_pid)
    vprint('*** Waiting for server to finish (5s)')
    isns_idle(5)
    return isns_start_server(server_config)

def isns_enroll_client(client_config, extra_args=None):
    """
    enroll client -- called for security mode
    """
    server = Global._isns_server_config_data[client_config.path]

    source_name = client_config.dict['SourceName']
    auth_name = client_config.dict.get('AuthName')
    auth_key = client_config.dict.get('AuthKeyFile')

    args = ['--local',
            '--enroll', auth_name,
            'node-name=%s' % source_name]
    if auth_key:
        args.append('key=%s.pub' % auth_key)
    if extra_args:
        args += extra_args

    (log_file, client_exit_value) = run_client(client_config, args)

    return client_exit_value

def run_client(client_config, args):
    """
    run a client instance -- arguments required
    """
    logfile = get_logfile_from_config(client_config.path)

    cmd = ['%s/isnsadm' % get_bindir(),
           '-c', client_config.path] + args
    exit_val = run_cmd(cmd, logfile)
    return (logfile, exit_val)

def isns_external_test(client_config, args):
    """
    Run an external test program
    """
    logfile = get_logfile_from_config(client_config.path)

    cmd = ['%s/%s' % (get_bindir(), args[0]),
           '-c', client_config.path] + args[1:]
    exit_val = run_cmd(cmd, logfile)
    return (logfile, exit_val)

def copy_file(from_file, to_file):
    dprint("Copying '%s' => '%s'" % (from_file, to_file))
    with open(from_file, 'r') as from_fd:
        with open(to_file, 'w') as to_fd:
            for line in from_fd:
                to_fd.write(line)

def skip_header(lines):
    #dprint("skip_header: first line: /%s/" % lines[0].rstrip())
    while lines:
        line = lines.pop(0)
        if line.startswith('-'):
            break
    #dprint("skip_header: now first line: /%s/" % lines[0].rstrip())

def get_next_db_object(lines):
    """
    From the input array of lines, return the glob of lines that make
    the next 'object'
    """
    tag_pat = re.compile('\s+([0-9a-fv]+)\s+')
    tags = []
    result = []
    while True:
        if not lines:
            break
        line = lines.pop(0)
        if line.startswith('-'):
            continue
        #dprint("looking at line: /%s/" % line)
        res = tag_pat.match(line)
        if res:
            if res.group(1) in Global._isns_ignore_tags:
                dprint("get_next_db_object: ignoring tag %s" % res.group(1))
                continue
            tags.append(line)
        else:
            if result:
                # we already have a result -- put this one back
                lines.insert(0, line)
                #dprint("break-ing -- we went to far!")
                break
            result.append(line)

    if tags:
        result += tags

    return result

def compare_db_objects(obj1, obj2):
    dprint("compare_db_objects")
    if obj1.size() != obj2.size():
        return False
    for i in range(obj1.size()):
        if obj1[i] != obj2[i]:
            return False
    return True

def load_dump(fname):
    lines = []
    dprint("load_dump(%s): entering" % fname)
    with open(fname) as fd:
        for line in fd:
            lines.append(line.rstrip())
    return lines

def verify_db_dump_file(data_file, dump_file):
    """
    Verify that the data (source) file and the
    dump (for comparison) files have the same 'objects'
    """
    dprint("verify_db_dump_file(%s, %s): entering" % (data_file, dump_file))

    data_lines = load_dump(data_file)
    dump_lines = load_dump(dump_file)

    skip_header(data_lines)
    skip_header(dump_lines)

    # I think this "line number" stuff is bogus, as it doesn't reflect any
    # actual line numbers
    line_no = 0

    while True:
            line_no += 1

            dump_obj = get_next_db_object(dump_lines)
            data_obj = get_next_db_object(data_lines)

            if not dump_obj and not data_obj:
                # both lists are empty -- done
                break

            if not dump_obj or not data_obj:
                vprint('*** %s: Excess data at end of dump or data' % \
                       Global._isns_stage_name)
                return (False, 'excess data in dump')

            #if not compare_db_objects(dump_obj, data_obj)

            if dump_obj != data_obj:
                vprint('*** Object mismatch (object %d)' % line_no)
                vprint('Expected:')
                vprint('%s' % '\n'.join(data_obj))
                vprint('Got:')
                vprint('%s' % '\n'.join(dump_obj))
                return (False, "Compare DB objects failed: Object at line %d" % line_no)

    if dump_lines:
        return (False, 'unexpected end of dump at line %d' % lin_no)

    return (True, '')

def verify_db(server_config):
    """
    Set pu to verify a database dump, then do the comparison
    """
    if not Global._isns_test_data:
        print("internal error: Test Case forgot to call set_up_test",
              file=sys.stderr)
        sys.exit(1)
    stage = Global._isns_stage_name
    dump_file = '%s/%s' % (Global._isns_test_dump, stage)

    dprint("planning to dump DB to %s" % dump_file)
    cmd = ['%s/isnsd' % get_bindir(),
           '-c', server_config.path,
           '--dump-db']
    cmd_result = run_cmd(cmd, dump_file)
    if cmd_result != 0:
        return (False, 'Could not dump DB (%d)' % cmd_result)

    data_dir = Global._isns_test_data
    data_file = '%s/%s' % (data_dir, stage)
    if not Global.security:
        data_file += '-no-security'

    # see if the reference data file exists. If it doesn't, this means
    # we are priming the test case, i.e. generating the data for the
    # first time, for future use, so just copy the file
    if not os.path.isfile(data_file):
        if not os.path.isdir(data_dir):
            os.makedirs(data_dir, 0o777)
        dprint("First Time: copying %s to %s" % (dump_file, data_file))
        copy_file(dump_file, data_file)
        return (False, 'Created initial data file: %s. Run again.' % data_file)

    # finally! verify the data matches
    vprint('*** Verifying database dump for stage % s' % stage)
    return verify_db_dump_file(data_file, dump_file)

def _isns_register_client(client_config, reg_params):
    args = ['--register']
    if reg_params:
        args += reg_params
    return run_client(client_config, args)

def isns_register_client(client_config, reg_params):
    """
    register a client
    """
    (logfile, exit_val) = _isns_register_client(client_config, reg_params)
    if exit_val == 0:
        return (True, '')
    return (False, 'failed to register client, logfile=%s' % logfile)


def _isns_query_objects(client_config, query_list):
    args = ['--query']
    if query_list:
        args += query_list
    return run_client(client_config, args)

def verify_query_response_file(data_file, dump_file):
    """
    verify that the query response files are the same
    """
    dprint("verify_query_response(%s, %s): entering" % (data_file, dump_file))

    data_lines = load_dump(data_file)
    dump_lines = load_dump(dump_file)

    stage = Global._isns_stage_name
    if len(dump_lines) > len(data_lines):
        vprint('*** %s: Excess data in dump at line %d' % (stage, len(data_lines) + 1))
        return (False, 'Excess data in dump at line: %d' % (len(data_lines) + 1))
    if len(dump_lines) < len(data_lines):
        vprint('*** %s: Unexpected end of dump at line %d' % (stage, len(dump_lines) + 1))
        return (False, 'Unexpected end of dump at line %d' % (len(dump_lines) + 1))

    start_pat = re.compile('\S')
    for line_no in range(len(dump_lines)):
        a = dump_lines[line_no]
        b = data_lines[line_no]
        if start_pat.match(a):
            if a == b:
                continue
            vprint('*** %s: Mismatch at line %d' % (stage, line_no))
            vprint('*** Found:    %s' % a)
            vprint('*** Expected: %s' % b)
            return (False, 'Mismatch at line %d' % line_no)

        (a_tag, a_value) = a.split()[1:3]
        (b_tag, b_value) = b.split()[1:3]

        if a_tag != b_tag:
            vprint('*** %s: Tag mismatch at line %d' % (stage, line_no))
            vprint('*** Found:    %s' % a_tag)
            vprint('*** Expected: %s' % b_tag)
            return (False, 'Tag mismatch at line %d' % line_no)

        if a_tag in Global._isns_ignore_tags:
            dprint("verify_query_response_file: ignoring tag %s" % res.group(1))
            continue

        if a_value != b_value:
            vprint('*** %s: Value mismatch at line %d' % (stage, line_no))
            vprint('*** Found:    %s' % a)
            vprint('*** Expected: %s' % b)
            return (False, 'Value mismatch at line %d' % line_no)

    return (True, '')

def verify_query_response(client_config):
    """
    Set up to verify a query response, then do the comparison
    """
    if not Global._isns_test_data:
        print("internal error: Test Case forgot to call set_up_test",
              file=sys.stderr)
        sys.exit(1)
    stage = Global._isns_stage_name

    dump_file = get_logfile_from_config(client_config.path)
    data_dir = Global._isns_test_data
    data_file = '%s/%s' % (data_dir, stage)
    if not Global.security:
        data_file += '-no-security'

    # see if the reference data file exists. If it doesn't, this means
    # we are priming the test case, i.e. generating the data for the
    # first time, for future use, so just copy the file
    if not os.path.isfile(data_file):
        if not os.path.isdir(data_dir):
            os.makedirs(data_dir, 0o777)
        dprint("First time: copying %s to %s" % (dump_file, data_file))
        copy_file(dump_file, data_file)
        return (False, 'Created initial data file: %s. Run again.' % data_file)

    # finally! verify data matches
    vprint('*** Verifying data for stage: %s ***' % stage)
    return verify_query_response_file(data_file, dump_file)

def isns_query_objects(client_config, query_list):
    """
    query object(s)
    """
    (logfile, exit_val) = _isns_query_objects(client_config, query_list)
    if exit_val != 0:
        return (False, 'query objects failed; logfile=%s' % logfile)
    return (True, '')

def isns_query_eid(client_config, extra_args=None):
    """
    Get the eid
    """
    args = ['--query-eid']
    if extra_args:
        args += extra_args

    vprint('*** Querying for EID ***')
    (logfile, exit_val) = run_client(client_config, args)
    if exit_val != 0:
        return (False, 0)

    # get the EID from the log file
    with open(logfile, 'r') as l_fd:
        eid_from_file = l_fd.readline().strip()
    dprint("Found EID=%s from %s!!!" % (eid_from_file, logfile))
    return (True, eid_from_file)

def isns_deregister_client(client_config, extra_args=None):
    """
    Deregister a client -- if no args are passed in then query and use the
    entity id (eid)
    """
    dprint('Deregister client, config=%s' % client_config.path)

    if not extra_args:
        (res, eid) = isns_query_eid(client_config)
        if not res:
            return (False, 'Cannot get our own EID!')
        extra_args = ['eid=%s' % eid]

    args = ['--deregister'] + extra_args

    (logfile, exit_val) = run_client(client_config, args)
    if exit_val != 0:
        return (False, 'deregister failed, exit val=%d' % exit_val)

    return (True, '')

def isns_register_domain(client_config, extra_args=None):
    """
    Register a discovery domain (DD)
    """
    args = ['--local', '--dd-register']
    if extra_args:
        args += extra_args

    (logfile, exit_val) = run_client(client_config, args)
    if exit_val != 0:
        return (False, 'dd-register failed, exit val=%d' % exit_val)

    return (True, '')

def isns_deregister_domain(client_config, extra_args):
    """
    Deregister a discovery domain (DD) -- extra_args are required, since we will
    always need a domain ID (the first argument)
    """
    args = ['--local', '--dd-deregister'] + extra_args
    (logfile, exit_val) = run_client(client_config, args)
    if exit_val != 0:
        return (False, 'dd-dregister (id=%s) failed, exit val=%d' % \
                (extra_args[0], exit_val))

    return (True, '')

def isns_finish():
    """
    all done!
    """
    dprint('isns_finish!')
    for pid in Global._isns_servers:
        dprint('Killing isnsd server pid=%d' % pid)
        try:
            os.kill(pid, 15)
        except ProcessLookupError:
            print('warning: cannot send signal to server pid=%d: no such process' % pid,
                  file=sys.stderr)
    Global._isns_servers = []



#!/usr/bin/env python3
"""
Unit tests for open-isns, using the unittest built-in package

********
WARNING:
	 Do NOT rename and of the 'stage names' (the first argument to isns_stage())
	withing a TestCase.

	Since the 'data/*' file names are based on the stage name, so changing
        one means you have to rename the data file(s) to match.
********

TODO:
    - Add option to 'populate' the 'data/*' directory tree, instead of failing
      each time a data file is not present, as well as creating the data file,
      which seems like a questionable side-effect methodology. This way, the only
      way to populate would be to pass this option. But it would not override
      existing data files (??).
    - Change the tests so that they verify the output directly from each step,
      rather than compare the output to previous runs, which is inherently buggy,
      since we may have saved bad data as a reference. So this method only
      really catches new bugs. (Enhancement)
    - Don't start the tests if an isnsd server is already running. Perhaps we
      should care about what port it is one? But if a daemon is left running by
      a previous failed test, new tests just mysteriously fail.
    - Add option to set the "__isns_bin_dir" in harness.py
    - Add a check in each subtest to ensure the previous subtest was run, so
      that trying to run just one random subtest will fail, even though allowed
      by unittest.
    - Get security working
"""

import sys
import unittest
import harness

class Test01(unittest.TestCase):
    """
    Simple test case: ensure a single client can talk to a server,
    including registering, querying, and deregistering, with and
    without security.
    """

    @classmethod
    def setUpClass(cls):
        harness.vprint('*** Starting %s' % cls.__name__)
        cls.security = harness.set_up_test(cls.__name__)
        cls.server_config = harness.create_server()
        cls.client_config = harness.create_client(cls.server_config)
        harness.isns_start_server(cls.server_config)

    def setUp(self):
        if harness.Global.debug or harness.Global.verbosity > 1:
            # this makes debug printing a little more clean
            print('', file=sys.stderr)

    def test01_enroll_client(self):
        """
        Enroll the test client
        """
        if self.security:
            harness.isns_stage('enroll', 'Enrolling client')
            self.assertEqual(harness.isns_enroll_client(self.client_config), 0)
            (res, msg) = harness.verify_db(self.server_config)
            self.assertTrue(res, msg)
            harness.vprint('*** SUCCESS ***')
        else:
            harness.vprint('*** Skipped (no security) ***')

    def test02_register_client_initiator_portal(self):
        """
        Register an iniator with default port
        """
        harness.isns_stage('registration', 'Register client initiator portal')
        (res, msg) = harness.isns_register_client(self.client_config,
                                                  ['initiator', 'portal'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    def test03_query_object_eid(self):
        """
        Run a simple query
        """
        harness.isns_stage('query', 'Querying eid')
        (res, msg) = harness.isns_query_objects(self.client_config, ['eid'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_query_response(self.client_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    def test99_deregister_client(self):
        """
        Deregister client
        """
        harness.isns_stage('deregister', 'Deregister client')
        (res, msg) = harness.isns_deregister_client(self.client_config)
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    @classmethod
    def tearDownClass(cls):
        harness.isns_finish()


class Test02(unittest.TestCase):
    """
    A little more complicated est case: ensure a two clients can
    talk to a server, including registering, queiying, and deregistering,
    with and without security.
    """

    @classmethod
    def setUpClass(cls):
        harness.vprint('*** Starting %s' % cls.__name__)
        cls.security = harness.set_up_test(cls.__name__)
        cls.server_config = harness.create_server()
        cls.client1_config = harness.create_client(cls.server_config, '127.1.0.1')
        cls.client2_config = harness.create_client(cls.server_config, '127.1.0.2')
        harness.isns_start_server(cls.server_config)

    def setUp(self):
        if harness.Global.debug or harness.Global.verbosity > 1:
            # this makes debug printing a little more clean
            print('', file=sys.stderr)

    def test01_enroll_clients(self):
        """
        Enroll the test client
        """
        if self.security:
            harness.isns_stage('enroll1', 'Enroll client 1')
            self.assertEqual(harness.isns_enroll_client(self.client1_config), 0)
            (res, msg) = harness.verify_db(self.server_config)
            self.assertTrue(res, msg)
            harness.isns_stage('enroll2', 'Enroll client 2')
            self.assertEqual(harness.isns_enroll_client(self.client2_config,
                                                        ['node-type=target']), 0)
            (res, msg) = harness.verify_db(self.server_config)
            self.assertTrue(res, msg)
            harness.vprint('*** SUCCESS ***')
        else:
            harness.vprint('*** Skipped (no security) ***')

    def test02_register_client_portals(self):
        """
        Register iniator and target portals
        """
        harness.isns_stage('registration1', 'Register client1: initiator portals')
        (res, msg) = harness.isns_register_client(self.client1_config,
                                                  ['initiator', 'portal'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.isns_stage('registration2', 'Register client2: target portals')
        (res, msg) = harness.isns_register_client(self.client2_config,
                                                  ['target', 'portal'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    def test03_query_object_eids_before(self):
        """
        Run a simple query on each client -- clients will not see each other (yet)
        """
        harness.isns_stage('query1-1', 'Querying client1 eid')
        (res, msg) = harness.isns_query_objects(self.client1_config, ['eid'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_query_response(self.client1_config)
        self.assertTrue(res, msg)
        harness.isns_stage('query1-2', 'Querying client2 eid')
        (res, msg) = harness.isns_query_objects(self.client2_config, ['eid'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_query_response(self.client2_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    def test04_register_discovery_domain(self):
        """
        Register a discovery domain (DD) that links the two clients
        """
        harness.isns_stage('dd-registration1', 'Add Discovery Domain')
        (res, msg) = harness.isns_register_domain(self.client1_config,
                                                  ['member-name=isns.client1',
                                                   'member-name=isns.client2'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    def test05_query_object_eids_after(self):
        """
        Run a simple query on each client -- clients should be able to see each
        other now
        """
        harness.isns_stage('query2-1', 'Querying client1 eid')
        (res, msg) = harness.isns_query_objects(self.client1_config, ['eid'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_query_response(self.client1_config)
        self.assertTrue(res, msg)
        harness.isns_stage('query2-2', 'Querying client2 eid')
        (res, msg) = harness.isns_query_objects(self.client2_config, ['eid'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_query_response(self.client2_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    def test06_query_object_node_type(self):
        """
        Have the first initiator query for the target
        """
        harness.isns_stage('query3', 'Querying client1 iscsi-node-type')
        (res, msg) = harness.isns_query_objects(self.client1_config, ['iscsi-node-type'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_query_response(self.client1_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    def test07_add_discovery_domain_member(self):
        """
        Have the first initiator query for the target
        """
        harness.isns_stage('dd-registration2', 'Add Discovery Domain Member')
        (res, msg) = harness.isns_register_domain(self.client1_config,
                                                  ['dd-id=1',
                                                   'member-name=isns.client2',
                                                   'member-name=iqn.com.foobar:disk1'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    def test08_query_discovery_domain_membersip(self):
        """
        Get a list of the Discovery Domains (DDs) we are a member of
        """
        harness.isns_stage('query4', 'Querying client1 dd-id')
        (res, msg) = harness.isns_query_objects(self.client1_config, ['dd-id'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_query_response(self.client1_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    def test09_remove_discovery_domain_entries(self):
        """
        Remove some entries from the DD (and add one)
        """
        harness.isns_stage('deregister1', 'Deregister Client by member-iscs-index')
        (res, msg) = harness.isns_deregister_domain(self.client1_config,
                                                    ['1', 'member-iscsi-idx=10'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)

        harness.isns_stage('deregister2', 'Deregister Client by member-name')
        (res, msg) = harness.isns_deregister_domain(self.client1_config,
                                                    ['1',
                                                     'member-name=iqn.com.foobar:disk1'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)

        harness.isns_stage('dd-registration3', 'Add Discovery Domain Member Again')
        (res, msg) = harness.isns_register_domain(self.client1_config,
                                                  ['member-name=isns.client2'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)

        harness.isns_stage('deregister3', 'Deregister Domaon')
        (res, msg) = harness.isns_deregister_domain(self.client1_config, ['1'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)

        harness.vprint('*** SUCCESS ***')

    @classmethod
    def tearDownClass(cls):
        harness.isns_finish()


class Test03(unittest.TestCase):
    """
    Validate registration and deregistration
    """

    @classmethod
    def setUpClass(cls):
        harness.vprint('*** Starting %s' % cls.__name__)
        cls.security = harness.set_up_test(cls.__name__)
        cls.server_config = harness.create_server()
        cls.client_config = harness.create_client(cls.server_config)
        harness.isns_start_server(cls.server_config)

    def setUp(self):
        if harness.Global.debug or harness.Global.verbosity > 1:
            # this makes debug printing a little more clean
            print('', file=sys.stderr)

    def test01_enroll_client(self):
        """
        Enroll our client
        """
        if self.security:
            harness.isns_stage('enroll', 'Enroll our cilent')
            self.assertEqual(harness.isns_enroll_client(self.client_config), 0)
            (res, msg) = harness.verify_db(self.server_config)
            self.assertTrue(res, msg)
            harness.vprint('*** SUCCESS ***')
        else:
            harness.vprint('*** Skipped (no security) ***')

    def test02_register_client_initiator_portal(self):
        """
        Register our client
        """
        harness.isns_stage('registration', 'Register client initiator portal')
        (res, msg) = harness.isns_register_client(self.client_config,
                                                  ['initiator', 'portal'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    def test03_deregister_client_portal(self):
        """
        Unregistering the portal should leave the iscsi node and
        portal group active, and move the portal to state limbo.
        """
        harness.isns_stage('deregister1', 'Deregister Client by portal')
        (res, msg) = harness.isns_deregister_client(self.client_config,
                                                    ['portal=127.0.0.1:860'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    def test04_deregister_client_name(self):
        """
        As the iscsi node goes away, so should the whole entity
        """
        harness.isns_stage('deregister2', 'Deregister Client by member-iscs-index')
        (res, msg) = harness.isns_deregister_client(self.client_config,
                                                    ['iscsi-name=isns.client1'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    @classmethod
    def tearDownClass(cls):
        harness.isns_finish()


class Test04(unittest.TestCase):
    """
    Verify database remains intact across server reboots
    """

    @classmethod
    def setUpClass(cls):
        harness.vprint('*** Starting %s' % cls.__name__)
        cls.security = harness.set_up_test(cls.__name__)
        cls.server_config = harness.create_server()
        cls.client_config = harness.create_client(cls.server_config)
        cls.server_pid = harness.isns_start_server(cls.server_config)

    def setUp(self):
        if harness.Global.debug or harness.Global.verbosity > 1:
            # this makes debug printing a little more clean
            print('', file=sys.stderr)

    def test01_enroll_client(self):
        """
        Enroll our client
        """
        if self.security:
            harness.isns_stage('enroll', 'Enroll our cilent')
            self.assertEqual(harness.isns_enroll_client(self.client_config), 0)
            (res, msg) = harness.verify_db(self.server_config)
            self.assertTrue(res, msg)
            harness.vprint('*** SUCCESS ***')
        else:
            harness.vprint('*** Skipped (no security) ***')

    def test02_register_client_initiator_portal(self):
        """
        Register our client
        """
        harness.isns_stage('registration', 'Register client initiator portal')
        (res, msg) = harness.isns_register_client(self.client_config,
                                                  ['initiator', 'portal'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    def test03_restart_server_verify_ok(self):
        """
        Restart the server and ensure the DB is still good
        """
        harness.isns_stage('restart', 'Restart the server and ensure DB ok')
        new_pid = harness.isns_restart_server(self.server_config, self.server_pid)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        self.assertNotEqual(self.server_pid, new_pid,
                            'New PID same as old? (%d)' % new_pid)
        harness.vprint('*** SUCCESS ***')

    @classmethod
    def tearDownClass(cls):
        harness.isns_finish()


class Test05(unittest.TestCase):
    """
    Test that an entity's registration expires as configured
    """

    @classmethod
    def setUpClass(cls):
        harness.vprint('*** Starting %s' % cls.__name__)
        cls.security = harness.set_up_test(cls.__name__)
        cls.server_config = harness.create_server()
        cls.client_config = harness.create_client(cls.server_config)
        cls.server_pid = harness.isns_start_server(cls.server_config)

    def setUp(self):
        if harness.Global.debug or harness.Global.verbosity > 1:
            # this makes debug printing a little more clean
            print('', file=sys.stderr)

    def test01_enroll_client(self):
        """
        Enroll our client
        """
        if self.security:
            harness.isns_stage('enroll', 'Enroll our cilent')
            self.assertEqual(harness.isns_enroll_client(self.client_config), 0)
            (res, msg) = harness.verify_db(self.server_config)
            self.assertTrue(res, msg)
            harness.vprint('*** SUCCESS ***')
        else:
            harness.vprint('*** Skipped (no security) ***')

    def test02_register_client_initiator_portal(self):
        """
        Register our client
        """
        harness.isns_stage('registration', 'Register client initiator portal')
        (res, msg) = harness.isns_register_client(self.client_config,
                                                  ['initiator', 'portal'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    def test03_verify_regitration_expires(self):
        """
        Ensure the registration expires as the 20 second registration period
        """
        harness.isns_stage('expired', 'Waiting for registration period to expire (25s)')
        harness.isns_idle(25)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    @classmethod
    def tearDownClass(cls):
        harness.isns_finish()


class Test06(unittest.TestCase):
    """
    Validate DevAttrReg replace mode
    """

    @classmethod
    def setUpClass(cls):
        harness.vprint('*** Starting %s' % cls.__name__)
        cls.security = harness.set_up_test(cls.__name__)
        cls.server_config = harness.create_server()
        cls.client_config = harness.create_client(cls.server_config)
        cls.server_pid = harness.isns_start_server(cls.server_config)

    def setUp(self):
        if harness.Global.debug or harness.Global.verbosity > 1:
            # this makes debug printing a little more clean
            print('', file=sys.stderr)

    def test01_enroll_client(self):
        """
        Enroll our client
        """
        if self.security:
            harness.isns_stage('enroll', 'Enroll our cilent')
            self.assertEqual(harness.isns_enroll_client(self.client_config), 0)
            (res, msg) = harness.verify_db(self.server_config)
            self.assertTrue(res, msg)
            harness.vprint('*** SUCCESS ***')
        else:
            harness.vprint('*** Skipped (no security) ***')

    def test02_register_client_initiator_portal(self):
        """
        Register our client
        """
        harness.isns_stage('registration1', 'Register client initiator portal')
        (res, msg) = harness.isns_register_client(self.client_config,
                                                  ['initiator', 'portal'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    def test03_get_client_eid(self):
        """
        Get our eid (entity id)
        """
        harness.isns_stage('query', 'Get our Entity ID (eid)')
        (res, eid) = harness.isns_query_eid(self.client_config)
        self.assertTrue(res, 'unable to get eid for client (%d)' % res)
        self.__class__.eid = eid
        harness.vprint('*** SUCCESS ***')

    def test04_replace_initiator_portal(self):
        """
        Replace the portal with different values
        """
        harness.isns_stage('registration2-1', 'Replace initiator portal')

        # set portal to 192.168.1.1, then override with 192.168.1.2
        (res, msg) = harness.isns_register_client(self.client_config,
                                                  ['--replace',
                                                   'entity=%s' % self.__class__.eid,
                                                   'initiator',
                                                   'portal=192.168.1.1:860'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        harness.isns_stage('registration2-2', 'Replace initiator portal')
        (res, msg) = harness.isns_register_client(self.client_config,
                                                  ['--replace',
                                                   'entity=%s' % self.__class__.eid,
                                                   'initiator',
                                                   'portal=192.168.1.2:860'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    def test05_register_discovery_domain(self):
        """
        Register our DD
        """
        harness.isns_stage('dd-registration1', 'Add Discovery Domain')
        (res, msg) = harness.isns_register_domain(self.client_config,
                                                  ['member-name=isns.client1'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    def test06_replace_initiator_portal_with_dd(self):
        """
        Replace the portal again. Now the object index of the initiator
        should not change, since it is a domain member now
        """
        harness.isns_stage('registration3', 'Replace initiator portal')
        (res, msg) = harness.isns_register_client(self.client_config,
                                                  ['--replace',
                                                   'entity=%s' % self.__class__.eid,
                                                   'initiator',
                                                   'portal=192.168.1.1:860'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    def test07_make_portal_dd_member(self):
        """
        Make the portal a domain member too. Now even the portal index should stay
        the same. Note that we do not replace whole entity now, but just the portal
        """
        harness.isns_stage('dd-registration2', 'Register port in Discovery Domain')
        (res, msg) = harness.isns_register_domain(self.client_config,
                                                  ['dd-id=1',
                                                   'member-addr=192.168.1.1',
                                                   'member-port=860'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.isns_stage('registration4', 'Replace whole entity')
        (res, msg) = harness.isns_register_client(self.client_config,
                                                  ['--replace',
                                                   '--key', 'portal=192.168.1.1:860',
                                                   'portal=192.168.1.2:860'])
        self.assertTrue(res, msg)
        harness.isns_stage('registration5', 'Replace whole entity again')
        (res, msg) = harness.isns_register_client(self.client_config,
                                                  ['--replace',
                                                   '--key', 'portal=192.168.1.2:860',
                                                   'portal=192.168.1.1:860'])
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    def test08_deregister_and_reregister(self):
        """
        Deregister the whole client, and re-register. Portal and Client
        index should remain the same
        """
        harness.isns_stage('deregister', 'Deregister our client')
        (res, msg) = harness.isns_deregister_client(self.client_config,
                                                    ['eid=%s' % self.__class__.eid])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.isns_stage('registration6', 'Re-Register the client')
        (res, msg) = harness.isns_register_client(self.client_config,
                                                  ['initiator',
                                                   'portal=192.168.1.1:860'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    @classmethod
    def tearDownClass(cls):
        harness.isns_finish()


class Test07(unittest.TestCase):
    """
    Validate that the server discards portals that do not
    respond to ESI messages
    """

    @classmethod
    def setUpClass(cls):
        harness.vprint('*** Starting %s' % cls.__name__)
        cls.security = harness.set_up_test(cls.__name__)
        cls.server_config = harness.create_server()
        cls.client_config = harness.create_client(cls.server_config)
        cls.server_pid = harness.isns_start_server(cls.server_config)

    def setUp(self):
        if harness.Global.debug or harness.Global.verbosity > 1:
            # this makes debug printing a little more clean
            print('', file=sys.stderr)

    def test01_enroll_client(self):
        """
        Enroll our client
        """
        if self.security:
            harness.isns_stage('enroll', 'Enroll our cilent')
            self.assertEqual(harness.isns_enroll_client(self.client_config), 0)
            (res, msg) = harness.verify_db(self.server_config)
            self.assertTrue(res, msg)
            harness.vprint('*** SUCCESS ***')
        else:
            harness.vprint('*** Skipped (no security) ***')

    def test02_register_client_initiator_portal(self):
        """
        Register a simple initiator with one portal, then wait for ESI to expire
        """
        harness.isns_stage('registration1', 'Register client initiator portal')
        (res, msg) = harness.isns_register_client(self.client_config,
                                                  ['initiator',
                                                   'portal,esi-port=65535,esi-interval=5'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.isns_stage('expired1', 'Wait for ESI to expire (15s)')
        harness.isns_idle(15)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    def test03_register_client_two_portals(self):
        """
        Register a simple initiator with two portals, one with ESI and one without.
        When the ESI-monitored portal expires, this should still take down
        the whole network entity.
        """
        harness.isns_stage('registration2', 'Register client with two portals')
        (res, msg) = harness.isns_register_client(self.client_config,
                                                  ['initiator',
                                                   'portal,esi-port=65535,esi-interval=5',
                                                   'portal=127.0.0.1:1'])
        harness.isns_stage('expired2', 'Wait for ESI to expire again (15s)')
        harness.isns_idle(15)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    @classmethod
    def tearDownClass(cls):
        harness.isns_finish()


class Test08(unittest.TestCase):
    """
    Validate using an external program

    Test case captured from a Wasabi Storage Builder registering itself

    For now, no security
    """

    @classmethod
    def setUpClass(cls):
        harness.vprint('*** Starting %s' % cls.__name__)
        cls.security = harness.set_up_test(cls.__name__, security=False)
        cls.server_config = harness.create_server()
        cls.client_config = harness.create_client(cls.server_config)
        cls.server_pid = harness.isns_start_server(cls.server_config)

    def setUp(self):
        if harness.Global.debug or harness.Global.verbosity > 1:
            # this makes debug printing a little more clean
            print('', file=sys.stderr)

    def test01_run_external_program_pauw1(self):
        """
        Run the first external program
        """
        harness.isns_stage('pauw1', 'Run external program pauw1')
        (res, msg) = harness.isns_external_test(self.client_config, ['pauw1'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    @classmethod
    def tearDownClass(cls):
        harness.isns_finish()


class Test09(unittest.TestCase):
    """
    Validate using an external program

    Test case captured from an iscsi-target registering itself

    For now, no security
    """

    @classmethod
    def setUpClass(cls):
        harness.vprint('*** Starting %s' % cls.__name__)
        cls.security = harness.set_up_test(cls.__name__, security=False)
        cls.server_config = harness.create_server()
        cls.client_config = harness.create_client(cls.server_config)
        cls.server_pid = harness.isns_start_server(cls.server_config)

    def setUp(self):
        if harness.Global.debug or harness.Global.verbosity > 1:
            # this makes debug printing a little more clean
            print('', file=sys.stderr)

    def test01_run_external_program_pauw2(self):
        """
        Run the second external program
        """
        harness.isns_stage('pauw2', 'Run external program pauw2')
        (res, msg) = harness.isns_external_test(self.client_config, ['pauw2'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    @classmethod
    def tearDownClass(cls):
        harness.isns_finish()


class Test10(unittest.TestCase):
    """
    Validate using an external program

    Test a problem reported by Albert, where re-registration shortly before
    ESI expiry fails to resurrect the registration properly

    Takes a while (20 seconds?)

    For now, no security
    """

    @classmethod
    def setUpClass(cls):
        harness.vprint('*** Starting %s' % cls.__name__)
        cls.security = harness.set_up_test(cls.__name__, security=False)
        cls.server_config = harness.create_server({'ESIMinInterval':'10s'})
        cls.client_config = harness.create_client(cls.server_config)
        cls.server_pid = harness.isns_start_server(cls.server_config)

    def setUp(self):
        if harness.Global.debug or harness.Global.verbosity > 1:
            # this makes debug printing a little more clean
            print('', file=sys.stderr)

    def test01_run_external_program_pauw3(self):
        """
        Run the third external program
        """
        harness.isns_stage('pauw3-1', 'Run external program pauw3 (slow)')
        (res, msg) = harness.isns_external_test(self.client_config,
                                                ['pauw3', '16'])
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    def test02_wait_for_esi_to_expire(self):
        """
        Wait for ESI to come around
        """
        harness.isns_stage('expired1', 'Wait for ESI to expire (5s)')
        harness.isns_idle(5)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    def test03_run_external_program_pauw3_again(self):
        """
        Run the third external program
        """
        harness.isns_stage('pauw3-2', 'Run external program pauw3 (slow)')
        (res, msg) = harness.isns_external_test(self.client_config,
                                                ['pauw3', '-n', '16'])
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    def test04_wait_for_esi_to_expire_again(self):
        """
        Wait for ESI to come around
        """
        harness.isns_stage('expired2', 'Wait for ESI to expire (5s)')
        harness.isns_idle(5)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    @classmethod
    def tearDownClass(cls):
        harness.isns_finish()


class Test11(unittest.TestCase):
    """
    Validate using an external program

    Test MS initiator registraion. The oddity about this is that the PG object
    preceeds the initiator object in the message.

    For now, no security
    """

    @classmethod
    def setUpClass(cls):
        harness.vprint('*** Starting %s' % cls.__name__)
        cls.security = harness.set_up_test(cls.__name__, security=False)
        cls.server_config = harness.create_server()
        cls.client_config = harness.create_client(cls.server_config)
        cls.server_pid = harness.isns_start_server(cls.server_config)

    def setUp(self):
        if harness.Global.debug or harness.Global.verbosity > 1:
            # this makes debug printing a little more clean
            print('', file=sys.stderr)

    def test01_run_external_program_pauw4(self):
        """
        Run the second external program
        """
        harness.isns_stage('pauw2', 'Run external program pauw4')
        (res, msg) = harness.isns_external_test(self.client_config, ['pauw4'])
        self.assertTrue(res, msg)
        (res, msg) = harness.verify_db(self.server_config)
        self.assertTrue(res, msg)
        harness.vprint('*** SUCCESS ***')

    @classmethod
    def tearDownClass(cls):
        harness.isns_finish()

# get a list of testcase classes, for listing
class_names_list=[]
for n in dir():
    if n.startswith("Test"):
        class_names_list.append(n)

def print_test_list_and_exit(verbosity):
    """
    print list of all tests

    always print the test cases, and print more if verbosity set

    verbosity       action
       0 (-1)       print list of test cases
       1 (default)  add in test case descriptions
       2 (-v)       add in Subtests and descriptions of them
    """
    global class_names_list

    for c in class_names_list:
        cls = getattr(sys.modules[__name__], c)
        print("*** Test Case: %s" % cls.__name__, end='')
        if verbosity > 0:
            print(cls.__doc__)
            if verbosity > 1:
                for m in dir(cls):
                    if m.startswith("test"):
                        method_doc = getattr(cls, m).__doc__
                        print("\t* Test Method: %s: %s" % (m, method_doc.strip()))
                print()
        else:
            print()
    sys.exit(0)

class MyTestProgram(unittest.TestProgram):
    """
    Augment TestProgram by adding a new method, which prints test case info
    """
    def list_tests(self):
        print_test_list_and_exit(harness.Global.verbosity)

if __name__ == '__main__':
    # do our own hackery first, to get access to verbosity, security, etc,
    # as well as add our own command-line options
    harness.setup_testProgram_overrides()
    # this next step parses args, gathers the tests, runs the tests, and displays
    # the results
    MyTestProgram()
